@@ -1,19 +1,21 @@
 from __future__ import unicode_literals
 from devpi_common.types import ensure_unicode
 from devpi_common.metadata import get_sorted_versions
+from devpi_common.validation import normalize_name
 from devpi_server.log import threadlog as log
-from devpi_web.doczip import iter_doc_contents
+from devpi_web.doczip import Docs
 import time
 
 
-def is_project_cached(stage, name):
+def is_project_cached(stage, project):
     if stage.ixconfig['type'] == 'mirror':
-        if not stage._load_project_cache(name):
+        if not stage.is_project_cached(project):
             return False
     return True
 
 
-def preprocess_project(stage, name):
+def preprocess_project(stage, name_input):
+    name = normalize_name(name_input)
     try:
         user = stage.user.name
         index = stage.index
@@ -34,7 +36,7 @@ def preprocess_project(stage, name):
         if links:
             # we assume it has been unpacked
             result['doc_version'] = version
-            result['+doczip'] = iter_doc_contents(stage, name, version)
+            result['+doczip'] = Docs(stage, name, version)
             break
         else:
             assert '+doczip' not in result
@@ -56,17 +58,15 @@ def iter_projects(xom):
         user_info = user.get(user)
         for index, index_info in user_info.get('indexes', {}).items():
             index = ensure_unicode(index)
-            stage = xom.model.getstage('%s/%s' % (username, index))
+            stage = xom.model.getstage(username, index)
             if stage is None:  # this is async, so the stage may be gone
                 continue
-            log.info("Search-Indexing %s/%s:" % (username, index))
-            names = stage.list_projectnames_perstage()
+            log.info("Search-Indexing %s:", stage.name)
+            names = stage.list_projects_perstage()
             for count, name in enumerate(names, start=1):
                 name = ensure_unicode(name)
                 current_time = time.time()
                 if current_time - timestamp > 3:
                     log.debug("currently search-indexed %s", count)
                     timestamp = current_time
-                if stage.get_projectname(name) is None:
-                    continue
                 yield preprocess_project(stage, name)
