@@ -1,10 +1,13 @@
 # coding: utf-8
 from __future__ import unicode_literals
+from defusedxml.xmlrpc import DefusedExpatParser
 from devpi_common.metadata import get_pyversion_filetype
 from devpi_common.metadata import get_sorted_versions
+from devpi_common.validation import normalize_name
 from devpi_common.viewhelp import iter_toxresults
 from devpi_server.log import threadlog as log
-from devpi_server.views import url_for_entrypath
+from devpi_server.readonly import SeqViewReadonly
+from devpi_server.views import StatusView, url_for_entrypath
 from devpi_web.description import get_description
 from devpi_web.doczip import Docs, get_unpack_path
 from devpi_web.indexing import is_project_cached
@@ -18,12 +21,19 @@ from pyramid.httpexceptions import HTTPBadGateway, HTTPError
 from pyramid.httpexceptions import HTTPFound, HTTPNotFound
 from pyramid.httpexceptions import default_exceptionresponse_view
 from pyramid.interfaces import IRoutesMapper
-from pyramid.response import FileResponse
+from pyramid.response import FileResponse, Response
 from pyramid.view import notfound_view_config, view_config
+from time import gmtime
+try:
+    from xmlrpc.client import Fault, Unmarshaller, dumps
+except ImportError:
+    from xmlrpclib import Fault, Unmarshaller, dumps
 import functools
 import json
 import py
 
+seq_types = (list, tuple, SeqViewReadonly)
+
 
 class ContextWrapper(object):
     def __init__(self, context):
@@ -34,31 +44,31 @@ class ContextWrapper(object):
 
     @reify
     def versions(self):
-        versions = self.stage.list_versions(self.name)
+        versions = self.stage.list_versions(self.project)
         if not versions:
-            raise HTTPNotFound("The project %s does not exist." % self.name)
+            raise HTTPNotFound("The project %s does not exist." % self.project)
         return get_sorted_versions(versions)
 
     @reify
     def stable_versions(self):
-        versions = self.stage.list_versions(self.name)
+        versions = self.stage.list_versions(self.project)
         if not versions:
-            raise HTTPNotFound("The project %s does not exist." % self.name)
+            raise HTTPNotFound("The project %s does not exist." % self.project)
         versions = get_sorted_versions(versions, stable=True)
         if not versions:
-            raise HTTPNotFound("The project %s has no stable release." % self.name)
+            raise HTTPNotFound("The project %s has no stable release." % self.project)
         return versions
 
     @reify
     def linkstore(self):
-        return self.stage.get_linkstore_perstage(self.name, self.version)
+        return self.stage.get_linkstore_perstage(self.project, self.version)
 
 
 def get_doc_info(context, request):
     relpath = request.matchdict['relpath']
     if not relpath:
         raise HTTPFound(location="index.html")
-    name = context.projectname
+    name = context.project
     version = context.version
     if version == 'latest':
         versions = context.versions
@@ -100,19 +110,19 @@ def doc_show(context, request):
     """ Shows the documentation wrapped in an iframe """
     context = ContextWrapper(context)
     stage = context.stage
-    name, version = context.projectname, context.version
+    name, version = context.project, context.version
     doc_info = get_doc_info(context, request)
     return dict(
         title="%s-%s Documentation" % (name, version),
         base_url=request.route_url(
             "docroot", user=stage.user.name, index=stage.index,
-            name=name, version=version, relpath=''),
+            project=name, version=version, relpath=''),
         baseview_url=request.route_url(
             "docviewroot", user=stage.user.name, index=stage.index,
-            name=name, version=version, relpath=''),
+            project=name, version=version, relpath=''),
         url=request.route_url(
             "docroot", user=stage.user.name, index=stage.index,
-            name=name, version=version, relpath=doc_info['relpath']),
+            project=name, version=version, relpath=doc_info['relpath']),
         version_mismatch=doc_info['version_mismatch'],
         doc_version=doc_info['doc_version'])
 
@@ -175,6 +185,17 @@ def format_timetuple(tt):
         return "{0}-{1:02d}-{2:02d} {3:02d}:{4:02d}:{5:02d}".format(*tt)
 
 
+def format_timestamp(ts, unset_value=None):
+    if ts is None:
+        ts = unset_value
+    try:
+        if ts is not None:
+            ts = format_timetuple(gmtime(ts)[:6])
+    except (ValueError, TypeError):
+        pass
+    return ts
+
+
 _what_map = dict(
     overwrite="Replaced",
     push="Pushed",
@@ -202,14 +223,14 @@ def get_files_info(request, linkstore, show_toxresults=False):
     filedata = linkstore.get_links(rel='releasefile')
     if not filedata:
         log.warn("project %r version %r has no files",
-                 linkstore.projectname, linkstore.version)
+                 linkstore.project, linkstore.version)
     for link in sorted(filedata, key=attrgetter('basename')):
         url = url_for_entrypath(request, link.entrypath)
         entry = link.entry
         if entry.eggfragment:
             url += "#egg=%s" % entry.eggfragment
-        elif entry.md5:
-            url += "#md5=%s" % entry.md5
+        elif entry.hash_spec:
+            url += "#" + entry.hash_spec
         py_version, file_type = get_pyversion_filetype(link.basename)
         if py_version == 'source':
             py_version = ''
@@ -227,7 +248,7 @@ def get_files_info(request, linkstore, show_toxresults=False):
             title=link.basename,
             url=url,
             basename=link.basename,
-            md5=entry.md5,
+            hash_spec=entry.hash_spec,
             dist_type=dist_file_types.get(file_type, ''),
             py_version=py_version,
             last_modified=last_modified,
@@ -270,16 +291,16 @@ def get_toxresults_info(linkstore, for_link, newest=True):
 
 
 def get_docs_info(request, stage, metadata):
-    if stage.name == 'root/pypi':
+    if stage.ixconfig['type'] == 'mirror':
         return
-    name, ver = metadata["name"], metadata["version"]
+    name, ver = normalize_name(metadata["name"]), metadata["version"]
     doc_path = get_unpack_path(stage, name, ver)
     if doc_path.exists():
         return dict(
             title="%s-%s" % (name, ver),
             url=request.route_url(
                 "docviewroot", user=stage.user.name, index=stage.index,
-                name=name, version=ver, relpath="index.html"))
+                project=name, version=ver, relpath="index.html"))
 
 
 @view_config(
@@ -295,11 +316,21 @@ def root(context, request):
         indexes = []
         for index in sorted(user.get('indexes', [])):
             stagename = "%s/%s" % (username, index)
+            stage = context.model.getstage(stagename)
             indexes.append(dict(
+                _ixconfig=stage.ixconfig,
                 title=stagename,
+                index_name=index,
+                index_title=stage.ixconfig.get('title', None),
+                index_description=stage.ixconfig.get('description', None),
                 url=request.stage_url(stagename)))
         users.append(dict(
+            _user=user,
             title=username,
+            user_name=username,
+            user_title=user.get('title', None),
+            user_description=user.get('description', None),
+            user_email=user.get('email', None),
             indexes=indexes))
     return dict(users=users)
 
@@ -313,16 +344,22 @@ def index_get(context, request):
     permissions = []
     bases = []
     packages = []
+    whitelist = []
     result = dict(
         title="%s index" % stage.name,
         simple_index_url=request.simpleindex_url(stage),
         permissions=permissions,
         bases=bases,
-        packages=packages)
-    if stage.name == "root/pypi":
+        packages=packages,
+        whitelist=whitelist,
+        index_name=stage.name,
+        index_title=stage.ixconfig.get('title', None),
+        index_description=stage.ixconfig.get('description', None))
+    if stage.ixconfig['type'] == 'mirror':
         return result
 
     if hasattr(stage, "ixconfig"):
+        whitelist.extend(sorted(stage.ixconfig['mirror_whitelist']))
         for base in stage.ixconfig["bases"]:
             bases.append(dict(
                 title=base,
@@ -350,43 +387,44 @@ def index_get(context, request):
                 special=special,
                 users=users))
 
-    for projectname in stage.list_projectnames_perstage():
-        version = stage.get_latest_version_perstage(projectname)
-        verdata = stage.get_versiondata_perstage(projectname, version)
+    for project in stage.list_projects_perstage():
+        version = stage.get_latest_version_perstage(project)
+        verdata = stage.get_versiondata_perstage(project, version)
         try:
-            name, ver = verdata["name"], verdata["version"]
+            name, ver = normalize_name(verdata["name"]), verdata["version"]
         except KeyError:
             log.error("metadata for project %r empty: %s, skipping",
-                      projectname, verdata)
+                      project, verdata)
             continue
-        show_toxresults = not (stage.user.name == 'root' and stage.index == 'pypi')
+        show_toxresults = (stage.ixconfig['type'] != 'mirror')
         linkstore = stage.get_linkstore_perstage(name, ver)
         packages.append(dict(
             info=dict(
                 title="%s-%s" % (name, ver),
                 url=request.route_url(
-                    "/{user}/{index}/{name}/{version}",
+                    "/{user}/{index}/{project}/{version}",
                     user=stage.user.name, index=stage.index,
-                    name=name, version=ver)),
+                    project=name, version=ver)),
             make_toxresults_url=functools.partial(
                 request.route_url, "toxresults",
                 user=stage.user.name, index=stage.index,
-                name=name, version=ver),
+                project=name, version=ver),
             files=get_files_info(request, linkstore, show_toxresults),
-            docs=get_docs_info(request, stage, verdata)))
+            docs=get_docs_info(request, stage, verdata),
+            _version_data=verdata))
     packages.sort(key=lambda x: x["info"]["title"])
 
     return result
 
 
 @view_config(
-    route_name="/{user}/{index}/{name}",
+    route_name="/{user}/{index}/{project}",
     accept="text/html", request_method="GET",
     renderer="templates/project.pt")
 def project_get(context, request):
     context = ContextWrapper(context)
     try:
-        releaselinks = context.stage.get_releaselinks(context.projectname)
+        releaselinks = context.stage.get_releaselinks(context.verified_project)
     except context.stage.UpstreamError as e:
         log.error(e.msg)
         raise HTTPBadGateway(e.msg)
@@ -394,8 +432,8 @@ def project_get(context, request):
     seen = set()
     for release in releaselinks:
         user, index = release.entrypath.split("/", 2)[:2]
-        name, version = release.projectname, release.version
-        if version == 'XXX':
+        name, version = release.project, release.version
+        if not version or version == 'XXX':
             continue
         seen_key = (user, index, name, version)
         if seen_key in seen:
@@ -405,23 +443,30 @@ def project_get(context, request):
             index_url=request.stage_url(user, index),
             title=version,
             url=request.route_url(
-                "/{user}/{index}/{name}/{version}",
-                user=user, index=index, name=name, version=version)))
+                "/{user}/{index}/{project}/{version}",
+                user=user, index=index, project=name, version=version),
+            _release=release))
         seen.add(seen_key)
+    if hasattr(context.stage, 'get_mirror_whitelist_info'):
+        whitelist_info = context.stage.get_mirror_whitelist_info(context.project)
+    else:
+        whitelist_info = dict(
+            has_mirror_base=context.stage.has_mirror_base(context.project),
+            blocked_by_mirror_whitelist=None)
     return dict(
-        title="%s/: %s versions" % (context.stage.name, name),
+        title="%s/: %s versions" % (context.stage.name, context.project),
+        blocked_by_mirror_whitelist=whitelist_info['blocked_by_mirror_whitelist'],
         versions=versions)
 
 
 @view_config(
-    route_name="/{user}/{index}/{name}/{version}",
+    route_name="/{user}/{index}/{project}/{version}",
     accept="text/html", request_method="GET",
     renderer="templates/version.pt")
 def version_get(context, request):
     """ Show version for the precise stage, ignores inheritance. """
     context = ContextWrapper(context)
-    user, index = context.username, context.index
-    name, version = context.name, context.version
+    name, version = context.verified_project, context.version
     stage = context.stage
     try:
         verdata = context.get_versiondata(perstage=True)
@@ -434,7 +479,7 @@ def version_get(context, request):
     for key, value in sorted(verdata.items()):
         if key in skipped_keys or key.startswith('+'):
             continue
-        if isinstance(value, list):
+        if isinstance(value, seq_types):
             if not len(value):
                 continue
             value = html.ul([html.li(x) for x in value]).unicode()
@@ -443,7 +488,7 @@ def version_get(context, request):
                 continue
             value = py.xml.escape(value)
         infos.append((py.xml.escape(key), value))
-    show_toxresults = not (user == 'root' and index == 'pypi')
+    show_toxresults = (stage.ixconfig['type'] != 'mirror')
     linkstore = stage.get_linkstore_perstage(name, version)
     files = get_files_info(request, linkstore, show_toxresults)
     docs = get_docs_info(request, stage, verdata)
@@ -460,12 +505,24 @@ def version_get(context, request):
     nav_links.append(dict(
         title="Simple index",
         url=request.route_url(
-            "/{user}/{index}/+simple/{name}",
-            user=context.username, index=context.index, name=context.name)))
-    if stage.has_pypi_base(name):
-        nav_links.append(dict(
-            title="PyPI page",
-            url="https://pypi.python.org/pypi/%s" % name))
+            "/{user}/{index}/+simple/{project}",
+            user=context.username, index=context.index, project=context.project)))
+    if hasattr(stage, 'get_mirror_whitelist_info'):
+        whitelist_info = stage.get_mirror_whitelist_info(name)
+    else:
+        whitelist_info = dict(
+            has_mirror_base=stage.has_mirror_base(name),
+            blocked_by_mirror_whitelist=False)
+    if whitelist_info['has_mirror_base']:
+        for base in reversed(list(stage.sro())):
+            if base.ixconfig["type"] != "mirror":
+                continue
+            mirror_web_url_fmt = base.ixconfig.get("mirror_web_url_fmt")
+            if not mirror_web_url_fmt:
+                continue
+            nav_links.append(dict(
+                title="%s page" % base.ixconfig.get("title", "Mirror"),
+                url=mirror_web_url_fmt.format(name=name)))
     return dict(
         title="%s/: %s-%s metadata and description" % (stage.name, name, version),
         content=get_description(stage, name, version),
@@ -473,15 +530,16 @@ def version_get(context, request):
         nav_links=nav_links,
         infos=infos,
         files=files,
+        blocked_by_mirror_whitelist=whitelist_info['blocked_by_mirror_whitelist'],
         show_toxresults=show_toxresults,
         make_toxresults_url=functools.partial(
             request.route_url, "toxresults",
             user=context.username, index=context.index,
-            name=context.name, version=context.version),
+            project=context.project, version=context.version),
         make_toxresult_url=functools.partial(
             request.route_url, "toxresult",
             user=context.username, index=context.index,
-            name=context.name, version=context.version))
+            project=context.project, version=context.version))
 
 
 @view_config(
@@ -496,12 +554,12 @@ def toxresults(context, request):
         linkstore, linkstore.get_links(basename=basename)[0], newest=False)
     return dict(
         title="%s/: %s-%s toxresults" % (
-            context.stage.name, context.name, context.version),
+            context.stage.name, context.project, context.version),
         toxresults=toxresults,
         make_toxresult_url=functools.partial(
             request.route_url, "toxresult",
             user=context.username, index=context.index,
-            name=context.name, version=context.version, basename=basename))
+            project=context.project, version=context.version, basename=basename))
 
 
 @view_config(
@@ -520,10 +578,63 @@ def toxresult(context, request):
         if x['basename'] == toxresult]
     return dict(
         title="%s/: %s-%s toxresult %s" % (
-            context.stage.name, context.name, context.version, toxresult),
+            context.stage.name, context.project, context.version, toxresult),
         toxresults=toxresults)
 
 
+@view_config(
+    route_name="/+status",
+    accept="text/html",
+    renderer="templates/status.pt")
+def statusview(request):
+    status = StatusView(request)._status()
+    replication_errors = []
+    for index, error in enumerate(status.get('replication-errors', {}).values()):
+        replication_errors.append(error)
+        if index >= 10:
+            replication_errors.append(dict(message="More than 10 replication errors."))
+    _polling_replicas = status.get('polling_replicas', {})
+    polling_replicas = []
+    for replica_uuid in sorted(_polling_replicas):
+        replica = _polling_replicas[replica_uuid]
+        polling_replicas.append(dict(
+            uuid=replica_uuid,
+            remote_ip=replica.get('remote-ip', 'unknown'),
+            outside_url=replica.get('outside-url', 'unknown'),
+            serial=replica.get('serial', 'unknown'),
+            in_request=replica.get('in-request', 'unknown'),
+            last_request=format_timestamp(
+                replica.get('last-request', 'unknown'))))
+    return dict(
+        msgs=request.status_info['msgs'],
+        info=dict(
+            uuid=status.get('uuid', 'unknown'),
+            role=status.get('role', 'unknown'),
+            outside_url=status.get('outside-url', 'unknown'),
+            master_url=status.get('master-url'),
+            master_uuid=status.get('master-uuid'),
+            master_serial=status.get('master-serial'),
+            master_serial_timestamp=format_timestamp(
+                status.get('master-serial-timestamp'), unset_value="never"),
+            replica_started_at=format_timestamp(
+                status.get('replica-started-at')),
+            replica_in_sync_at=format_timestamp(
+                status.get('replica-in-sync-at'), unset_value="never"),
+            update_from_master_at=format_timestamp(
+                status.get('update-from-master-at'), unset_value="never"),
+            serial=status.get('serial', 'unknown'),
+            last_commit_timestamp=format_timestamp(
+                status.get('last-commit-timestamp', 'unknown')),
+            event_serial=status.get('event-serial', 'unknown'),
+            event_serial_timestamp=format_timestamp(
+                status.get('event-serial-timestamp', 'unknown')),
+            event_serial_in_sync_at=format_timestamp(
+                status.get('event-serial-in-sync-at', 'unknown'),
+                unset_value="never")),
+        replication_errors=replication_errors,
+        polling_replicas=polling_replicas)
+
+
 def batch_list(num, current, left=3, right=3):
     result = []
     if not num:
@@ -652,15 +763,25 @@ class SearchView:
                 continue
             elif text_type in ('title', 'page'):
                 docs = self.get_docs(stage, data)
-                entry = docs[sub_data['text_path']]
-                text = entry['text']
-                highlight = search_index.highlight(text, sub_hit.get('words'))
+                try:
+                    entry = docs[sub_data['text_path']]
+                except KeyError:
+                    highlight = (
+                        "Couldn't access documentation files for %s "
+                        "version %s on %s. This is a bug. If you find a way "
+                        "to reproduce this, please file an issue at: "
+                        "https://bitbucket.org/hpk42/devpi/issues" % (
+                            data['name'], data['doc_version'], stage.name))
+                else:
+                    text = entry['text']
+                    highlight = search_index.highlight(text, sub_hit.get('words'))
                 title = sub_data.get('text_title', title)
                 text_path = sub_data.get('text_path')
                 if text_path:
                     sub_hit['url'] = self.request.route_url(
                         "docviewroot", user=data['user'], index=data['index'],
-                        name=data['name'], version=data['doc_version'],
+                        project=normalize_name(data['name']),
+                        version=data['doc_version'],
                         relpath="%s.html" % text_path)
             elif text_type in ('keywords', 'description', 'summary'):
                 metadata = self.get_versiondata(stage, data)
@@ -672,9 +793,10 @@ class SearchView:
                 highlight = search_index.highlight(text, sub_hit.get('words'))
                 if 'version' in data:
                     sub_hit['url'] = self.request.route_url(
-                        "/{user}/{index}/{name}/{version}",
+                        "/{user}/{index}/{project}/{version}",
                         user=data['user'], index=data['index'],
-                        name=data['name'], version=data['version'],
+                        project=normalize_name(data['name']),
+                        version=data['version'],
                         _anchor=text_type)
             else:
                 log.error("Unknown type %s" % text_type)
@@ -697,14 +819,16 @@ class SearchView:
                 continue
             if 'version' in data:
                 item['url'] = self.request.route_url(
-                    "/{user}/{index}/{name}/{version}",
+                    "/{user}/{index}/{project}/{version}",
                     user=data['user'], index=data['index'],
-                    name=data['name'], version=data['version'])
+                    project=normalize_name(data['name']),
+                    version=data['version'])
                 item['title'] = "%s-%s" % (data['name'], data['version'])
             else:
                 item['url'] = self.request.route_url(
-                    "/{user}/{index}/{name}",
-                    user=data['user'], index=data['index'], name=data['name'])
+                    "/{user}/{index}/{project}",
+                    user=data['user'], index=data['index'],
+                    project=normalize_name(data['name']))
                 item['title'] = data['name']
             item['sub_hits'] = self.process_sub_hits(
                 stage, item['sub_hits'], data)
@@ -737,3 +861,94 @@ class SearchView:
         renderer='templates/search_help.pt')
     def search_help(self):
         return dict()
+
+    def query_from_xmlrpc(self, body):
+        unmarshaller = Unmarshaller()
+        parser = DefusedExpatParser(unmarshaller)
+        parser.feed(body)
+        parser.close()
+        (data, method) = (unmarshaller.close(), unmarshaller.getmethodname())
+        if method != "search":
+            raise ValueError("Unknown method '%s'." % method)
+        if len(data) == 2:
+            query, operator = data
+        else:
+            query = data
+            operator = "and"
+        log.debug("xmlrpc_search {0}".format((query, operator)))
+        operator = operator.upper()
+        if operator not in ('AND', 'OR', 'ANDNOT', 'ANDMAYBE', 'NOT'):
+            raise ValueError("Unknown operator '%s'." % operator)
+        if set(query.keys()).difference(['name', 'summary']):
+            raise ValueError("Only 'name' and 'summary' allowed in query.")
+        parts = []
+        for key, field in (('name', 'project'), ('summary', 'summary')):
+            value = query.get(key, [])
+            if len(value) == 0:
+                continue
+            elif len(value) == 1:
+                parts.append('(type:%s "%s")' % (field, value[0].replace('"', '')))
+            else:
+                raise ValueError("Only on value allowed for query.")
+        return (" %s " % operator).join(parts)
+
+    def search_index_packages(self, query):
+        search_index = self.request.registry['search_index']
+        result = search_index.query_projects(query, page=None)
+        context = ContextWrapper(self.request.context)
+        sro = dict((x.name, i) for i, x in enumerate(context.stage.sro()))
+        # first gather basic info and only get most relevant info based on
+        # stage resolution order
+        name2stage = {}
+        name2data = {}
+        for item in result['items']:
+            data = item['data']
+            path = data['path']
+            stage = self.get_stage(path)
+            if stage is None:
+                continue
+            if stage.name not in sro:
+                continue
+            name = data['name']
+            current_stage = name2stage.get(name)
+            if current_stage is None or sro[stage.name] < sro[current_stage.name]:
+                name2stage[name] = stage
+                try:
+                    score = item['score']
+                except KeyError:
+                    score = False
+                    sub_hits = item.get('sub_hits')
+                    if sub_hits:
+                        score = sub_hits[0].get('score', False)
+                name2data[name] = dict(
+                    version=data.get('version', ''),
+                    score=score)
+        # then gather more info if available and build results
+        hits = []
+        for name, stage in name2stage.items():
+            data = name2data[name]
+            version = data['version']
+            if version and is_project_cached(stage, name):
+                metadata = stage.get_versiondata(name, version)
+                version = metadata.get('version', version)
+                summary = metadata.get('summary', '')
+            else:
+                summary = ''
+            hits.append(dict(
+                name="/%s/%s" % (stage.name, name), summary=summary,
+                version=version, _pypi_ordering=data['score']))
+        return hits
+
+    @view_config(
+        route_name="/{user}/{index}/", request_method="POST",
+        content_type="text/xml")
+    def xmlrpc_search(self):
+        try:
+            query = self.query_from_xmlrpc(self.request.body)
+            log.debug("xmlrpc_search {0}".format(query))
+            hits = self.search_index_packages(query)
+            response = dumps((hits,), methodresponse=1, encoding='utf-8')
+        except Exception as e:
+            log.exception("Error in xmlrpc_search")
+            response = dumps(Fault(1, repr(e)), encoding='utf-8')
+        return Response(response)
