@@ -2,7 +2,8 @@ from test_devpi_server.conftest import gentmp, httpget, makemapp  # noqa
 from test_devpi_server.conftest import maketestapp, makexom, mapp  # noqa
 from test_devpi_server.conftest import pypiurls, testapp, pypistage  # noqa
 from test_devpi_server.conftest import dummyrequest, pypiurls, testapp  # noqa
-from test_devpi_server.conftest import mock  # noqa
+from test_devpi_server.conftest import storage_info  # noqa
+from test_devpi_server.conftest import mock, pyramidconfig  # noqa
 import pytest
 
 
